/*************************************************************
 * File  :  hello_rtsa.c
 * Module:  Agora SD-RTN SDK RTC C API demo application.
 *
 * This is a part of the Agora RTC Service SDK.
 * Copyright (C) 2020 Agora IO
 * All rights reserved.
 *
 *************************************************************/

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "agora_rtc_api.h"
#include "file_parser.h"
#include "file_writer.h"
#include "utility.h"
#include "pacer.h"
#include "log.h"

#define DEFAULT_CHANNEL_NAME "hello_demo"
#define DEFAULT_CERTIFACTE_FILENAME "certificate.bin"
#define DEFAULT_CREDENTIAL_FILENAME "deviceID.bin"
#define DEFAULT_SEND_VIDEO_FILENAME "send_video.h264"
#define DEFAULT_SEND_AUDIO_FILENAME "send_audio_16k_1ch.pcm"

#define DEFAULT_RECV_AUDIO_BASENAME "recv_audio"
#define DEFAULT_RECV_VIDEO_BASENAME "recv_video"
#define DEFAULT_SEND_VIDEO_FRAME_RATE (25)
#define DEFAULT_BANDWIDTH_ESTIMATE_MIN_BITRATE (100000)
#define DEFAULT_BANDWIDTH_ESTIMATE_MAX_BITRATE (1000000)
#define DEFAULT_BANDWIDTH_ESTIMATE_START_BITRATE (500000)
#define DEFAULT_SEND_AUDIO_FRAME_DURATION_MS (20)
#define DEFAULT_PCM_SAMPLE_RATE (16000)
#define DEFAULT_PCM_CHANNEL_NUM (1)

typedef struct {
	const char *p_sdk_log_dir;

	const char *p_appid;
	const char *p_token;
	const char *p_channel;
	uint32_t uid;

	// video related config
	int send_video_type;
	int send_video_frame_rate;
	const char *send_video_file_path;

	// audio related config
	int send_audio_type;
	const char *send_audio_file_path;

	uint32_t pcm_sample_rate;
	uint32_t pcm_channel_num;

	// advanced config
	bool send_video_generic_flag;
} app_config_t;

typedef struct {
	app_config_t config;

	void *video_file_parser;
	void *video_file_writer;

	void *audio_file_parser;
	void *audio_file_writer;

	bool b_stop_flag;
	bool b_connected_flag;
} app_t;

static app_t g_app = {
    .config = {
        // common config
        .p_sdk_log_dir              =   "io.agora.rtc_sdk",
        .p_appid                    =   "",
        .p_channel                  =   DEFAULT_CHANNEL_NAME,
        .p_token                    =   "",
        .uid                        =   0,

        // video related config
        .send_video_type            =   VIDEO_CODEC_H264,
        .send_video_frame_rate      =   DEFAULT_SEND_VIDEO_FRAME_RATE,
        .send_video_file_path       =   DEFAULT_SEND_VIDEO_FILENAME,

        // audio related config
        .send_audio_type            =   AUDIO_CODEC_PCM,
        .send_audio_file_path       =   DEFAULT_SEND_AUDIO_FILENAME,

        // pcm related config
        .pcm_sample_rate            =   DEFAULT_PCM_SAMPLE_RATE,
        .pcm_channel_num            =   DEFAULT_PCM_CHANNEL_NUM,

        // advanced config
        .send_video_generic_flag    =   false,
    },

    .video_file_parser      = NULL,
    .video_file_writer      = NULL,

    .audio_file_parser      = NULL,
    .audio_file_writer      = NULL,

    .b_stop_flag            = false,
    .b_connected_flag       = false,
};

static void app_signal_handler(int sig)
{
	switch (sig) {
	case SIGQUIT:
	case SIGABRT:
	case SIGINT:
		g_app.b_stop_flag = true;
		break;
	default:
		LOGW("no handler, sig=%d", sig);
	}
}

void app_print_usage(int argc, char **argv)
{
	LOGS("\nUsage: %s [OPTION]", argv[0]);
	LOGS(" -h, --help                : show help info");
	LOGS(" -i, --app-id              : application id, either app-id OR token MUST be set");
	LOGS(" -t, --token               : token for authentication");
	LOGS(" -c, --channel-id          : channel, default is 'demo'");
	LOGS(" -u, --user-id             : user id, default is 0");
	LOGS(" -f  --fps                 : video frame rate, default is 30");
	LOGS(" -v, --video-codec         : video codec, default is 2");
	LOGS("                             support: 2=H264, 20=JPEG");
	LOGS(" -a, --audio-codec         : audio codec, default is 100");
	LOGS("                             support: 1=OPUS, 5=G722, 8=AACLC, 9=HEAAC, 100=PCM");
	LOGS(" -l, --log-dir             : agora SDK log directory, default is '.'");
	LOGS(" -s, --send-video-file     : send video file path, default is './%s'",
		 DEFAULT_SEND_VIDEO_FILENAME);
	LOGS(" -S, --send-audio-file     : send audio file path, default is './%s'",
		 DEFAULT_SEND_AUDIO_FILENAME);
	LOGS(" -m, --pcm-sample-rate     : sample rate for the sending and receiving pcm stream");
	LOGS(" -n, --pcm-channel-num     : channel number for the sending and receiving pcm stream");
	LOGS(" -g, --send-video-generic  : enable generic codec flag for sending video");

	LOGS("\nExample:");
	LOGS("    %s --app-id xxx [--token xxx] --channel-id xxx --send-video-file ./video.h264 --fps 15 \
--send-audio-file ./audio.aac",
		 argv[0]);
}

int app_parse_args(int argc, char **argv)
{
	app_config_t *config = &g_app.config;
	const char *av_short_option = "hi:t:c:u:v:f:a:S:s:m:n:g";
	const struct option av_long_option[] = { { "help", 0, NULL, 'h' },
											 { "app-id", 1, NULL, 'i' },
											 { "token", 1, NULL, 't' },
											 { "channel-id", 1, NULL, 'c' },
											 { "user-id", 1, NULL, 'u' },
											 { "audio-codec", 1, NULL, 'a' },
											 { "video-codec", 1, NULL, 'v' },
											 { "fps", 1, NULL, 'f' },
											 { "send-audio-file", 1, NULL, 'S' },
											 { "send-video-file", 1, NULL, 's' },
											 { "pcm-sample-rate", 1, NULL, 'm' },
											 { "pcm-channel-num", 1, NULL, 'n' },
											 { "send-video-generic", 0, NULL, 'g' },
											 { 0, 0, 0, 0 } };

	int ch = -1;
	int optidx = 0;
	int rval = 0;

	while (1) {
		ch = getopt_long(argc, argv, av_short_option, av_long_option, &optidx);
		if (ch == -1) {
			break;
		}

		switch (ch) {
		case 'h':
			return -1;
		case 'i':
			config->p_appid = optarg;
			break;
		case 't':
			config->p_token = optarg;
			break;
		case 'c':
			config->p_channel = optarg;
			break;
		case 'u':
			config->uid = strtol(optarg, NULL, 10);
			break;
		case 'v':
			config->send_video_type = strtol(optarg, NULL, 10);
			break;
		case 'f':
			config->send_video_frame_rate = strtol(optarg, NULL, 10);
			break;
		case 'a':
			config->send_audio_type = strtol(optarg, NULL, 10);
			break;
		case 'S':
			config->send_audio_file_path = optarg;
			break;
		case 's':
			config->send_video_file_path = optarg;
			break;
		case 'g':
			config->send_video_generic_flag = true;
			break;
		case 'm':
			config->pcm_sample_rate = atoi(optarg);
			break;
		case 'n':
			config->pcm_channel_num = atoi(optarg);
			break;
		default:
			LOGS("Unknown cmd param %s", av_long_option[optidx].name);
			return -1;
		}
	}

	// check parameter sanity
	if (strcmp(config->p_appid, "") == 0) {
		LOGE("MUST provide App ID");
		return -1;
	}

	if (strcmp(config->p_channel, "") == 0) {
		LOGE("MUST provide channel name");
		return -1;
	}

	if (config->send_video_frame_rate <= 0) {
		LOGE("Invalid video frame rate: %d", config->send_video_frame_rate);
		return -1;
	}

	return 0;
}

static int app_init(void)
{
	app_config_t *config = &g_app.config;
	parser_cfg_t parser_cfg = { 0 };

	signal(SIGQUIT, app_signal_handler);
	signal(SIGABRT, app_signal_handler);
	signal(SIGINT, app_signal_handler);

	if (config->send_audio_type == AUDIO_CODEC_PCM) {
		parser_cfg.u.audio_cfg.sampleRateHz = config->pcm_sample_rate;
		parser_cfg.u.audio_cfg.numberOfChannels = config->pcm_channel_num;
	}

	g_app.video_file_parser =
			create_file_parser(config->send_video_type, config->send_video_file_path, NULL);
	if (!g_app.video_file_parser) {
		return -1;
	}

	g_app.video_file_writer = create_file_writer(DEFAULT_RECV_VIDEO_BASENAME);

	if (config->send_video_generic_flag) {
		config->send_video_type = VIDEO_CODEC_GENERIC;
	}

	g_app.audio_file_parser =
			create_file_parser(config->send_audio_type, config->send_audio_file_path, &parser_cfg);
	if (!g_app.audio_file_parser) {
		return -1;
	}

	g_app.audio_file_writer = create_file_writer(DEFAULT_RECV_AUDIO_BASENAME);

	return 0;
}

static int app_send_audio(void)
{
	app_config_t *config = &g_app.config;

	frame_t frame;
	if (file_parser_obtain_frame(g_app.audio_file_parser, &frame) < 0) {
		LOGE("The file parser failed to obtain audio frame");
		return -1;
	}

	// API: send audio data
	audio_frame_info_t info = { 0 };
	info.codec = config->send_audio_type;
	int rval = agora_rtc_send_audio_data(config->p_channel, frame.ptr, frame.len, &info);
	if (rval < 0) {
		LOGE("Failed to send audio data, reason: %s", agora_rtc_err_2_str(rval));
		return -1;
	}

	return 0;
}

static int app_send_video(void)
{
	app_config_t *config = &g_app.config;
	uint8_t stream_id = 0;

	frame_t frame;
	if (file_parser_obtain_frame(g_app.video_file_parser, &frame) < 0) {
		LOGE("The file parser failed to obtain audio frame");
		return -1;
	}

	// API: send vido data
	video_frame_info_t info;
	info.type = frame.u.video.key_frame_flag ? VIDEO_FRAME_KEY : VIDEO_FRAME_DELTA;
	info.frames_per_sec = config->send_video_frame_rate;
	info.codec = config->send_video_type;
	int rval = agora_rtc_send_video_data(config->p_channel, stream_id, frame.ptr, frame.len, &info);
	if (rval < 0) {
		LOGE("Failed to send video data, reason: %s", agora_rtc_err_2_str(rval));
		return -1;
	}

	return 0;
}

static void __on_join_channel_success(const char *channel, int elapsed)
{
	g_app.b_connected_flag = true;
	LOGI("Join the channel \"%s\" successfully, elapsed %d ms", channel, elapsed);
}

static void __on_connection_lost(const char *channel)
{
	g_app.b_connected_flag = false;
	LOGW("Lost connection from the channel \"%s\"", channel);
}

static void __on_rejoin_channel_success(const char *channel, int elapsed_ms)
{
	g_app.b_connected_flag = true;
	LOGI("Rejoin the channel \"%s\" successfully, elapsed %d ms", channel, elapsed_ms);
}

static void __on_user_joined(const char *channel, uint32_t uid, int elapsed_ms)
{
	LOGI("Remote user \"%u\" has joined the channel \"%s\", elapsed %d ms", uid, channel,
		 elapsed_ms);
}

static void __on_user_offline(const char *channel, uint32_t uid, int reason)
{
	LOGI("Remote user \"%u\" has left the channel \"%s\", reason %d", uid, channel, reason);
}

static void __on_error(const char *channel, int code, const char *msg)
{
	if (code == ERR_INVALID_APP_ID) {
		LOGE("Invalid App ID. Please double check. Error msg \"%s\"", msg);
		g_app.b_stop_flag = true;
	} else if (code == ERR_INVALID_CHANNEL_NAME) {
		LOGE("Invalid channel name \"%s\". Please double check. Error msg \"%s\"", channel, msg);
		g_app.b_stop_flag = true;
	} else if (code == ERR_INVALID_TOKEN || code == ERR_TOKEN_EXPIRED) {
		LOGE("Invalid token. Please double check. Error msg \"%s\"", msg);
		g_app.b_stop_flag = true;
	} else if (code == ERR_SEND_VIDEO_OVER_BANDWIDTH_LIMIT) {
		LOGE("Not enough uplink bandwdith. Error msg \"%s\"", msg);
	} else {
		LOGW("Error %d is captured. Error msg \"%s\"", code, msg);
	}
}

static void __on_audio_data(const char *channel, const uint32_t uid, uint16_t sent_ts,
							audio_codec_e codec, const void *data, size_t len)
{
	//LOGD("on_audio_data, uid %u, sent_ts %d, codec %d, len %zu", uid, sent_ts, codec, len);
	write_file(g_app.audio_file_writer, codec, data, len);
}

static void __on_video_data(const char *channel, const uint32_t uid, uint16_t sent_ts,
							video_codec_e codec, uint8_t stream_id, int is_key_frame,
							const void *data, size_t len)
{
	//LOGD("on_video_data: uid %u, key_frame %d, codec %u, len %zu", uid, is_key_frame, codec, len);
	write_file(g_app.video_file_writer, codec, data, len);
}

static void __on_target_bitrate_changed(const char *channel, uint32_t target_bps)
{
	LOGI("Bandwidth change detected. Please adjust encoder bitrate to %u kbps", target_bps / 1000);
}

static void __on_key_frame_gen_req(const char *channel, uint32_t uid, uint8_t stream_id)
{
	LOGW("Frame loss detected. Please notify the encoder to generate key frame immediately");
}

static agora_rtc_event_handler_t event_handler = {
	.on_join_channel_success = __on_join_channel_success,
	.on_connection_lost = __on_connection_lost,
	.on_rejoin_channel_success = __on_rejoin_channel_success,
	.on_user_joined = __on_user_joined,
	.on_user_offline = __on_user_offline,
	.on_target_bitrate_changed = __on_target_bitrate_changed,
	.on_key_frame_gen_req = __on_key_frame_gen_req,
	.on_video_data = __on_video_data,
	.on_error = __on_error,
	.on_audio_data = __on_audio_data,
};

int main(int argc, char **argv)
{
	app_config_t *config = &g_app.config;
	int rval;

	// 1. app parse args
	rval = app_parse_args(argc, argv);
	if (rval < 0) {
		app_print_usage(argc, argv);
		return -1;
	}

	LOGI("Welcome to RTSA SDK v%s", agora_rtc_get_version());

	// 2. app init
	rval = app_init();
	if (rval < 0) {
		LOGE("Failed to initialize application");
		return -1;
	}

	// 3. API: verify license
	char *str_certificate = util_get_string_from_file(DEFAULT_CERTIFACTE_FILENAME);
	char *str_credetial = util_get_string_from_file(DEFAULT_CREDENTIAL_FILENAME);
	if (str_certificate && str_credetial) {
		rval = agora_rtc_license_verify(str_certificate, strlen(str_certificate), str_credetial,
										strlen(str_credetial));
		if (rval < 0) {
			LOGE("Failed to verify license, reason: %s", agora_rtc_err_2_str(rval));
			return -1;
		}
		LOGI("Verify license successfully");
	}

	// 4. API: init agora rtc sdk
	int appid_len = strlen(config->p_appid);
	void *p_appid = (void *)(appid_len == 0 ? NULL : config->p_appid);
	rval = agora_rtc_init(p_appid, config->uid, &event_handler, config->p_sdk_log_dir);
	if (rval < 0) {
		LOGE("Failed to initialize Agora sdk, reason: %s", agora_rtc_err_2_str(rval));
		return -1;
	}

	agora_rtc_set_log_level(RTC_LOG_NOTICE);
	agora_rtc_set_bwe_param(DEFAULT_BANDWIDTH_ESTIMATE_MIN_BITRATE,
							DEFAULT_BANDWIDTH_ESTIMATE_MAX_BITRATE,
							DEFAULT_BANDWIDTH_ESTIMATE_START_BITRATE);

	// 5. API: join channel
	int token_len = strlen(config->p_token);
	void *p_token = (void *)(token_len == 0 ? NULL : config->p_token);

	rtc_channel_options_t channel_options = { 0 };
	channel_options.auto_subscribe_audio = true;
	channel_options.auto_subscribe_video = true;

	if (config->send_audio_type == AUDIO_CODEC_PCM) {
		/* If we want to send PCM data instead of encoded audio like AAC or Opus, here please enable
           audio codec, as well as configure the PCM sample rate and number of channels */
		channel_options.audio_codec_opt.use_audio_codec = true;
		channel_options.audio_codec_opt.pcm_sample_rate = config->pcm_sample_rate;
		channel_options.audio_codec_opt.pcm_channel_num = config->pcm_channel_num;
	}

	rval = agora_rtc_join_channel(config->p_channel, p_token, token_len, &channel_options);
	if (rval < 0) {
		LOGE("Failed to join channel \"%s\", reason: %s", config->p_channel,
			 agora_rtc_err_2_str(rval));
		return -1;
	}

	// 7. send video and audio in a loop
	int audio_send_interval_ms = DEFAULT_SEND_AUDIO_FRAME_DURATION_MS;
	int video_send_interval_ms = 1000 / config->send_video_frame_rate;
	void *pacer = pacer_create(audio_send_interval_ms, video_send_interval_ms);

	while (!g_app.b_stop_flag) {
		if (g_app.b_connected_flag && is_time_to_send_audio(pacer)) {
			app_send_audio();
		}
		if (g_app.b_connected_flag && is_time_to_send_video(pacer)) {
			app_send_video();
		}
		// sleep and wait until time is up for next send
		wait_before_next_send(pacer);
	}

	// 8. API: leave channel
	agora_rtc_leave_channel(config->p_channel);

	// 9. API: fini rtc sdk
	agora_rtc_fini();

	// 10. app clean up
	pacer_destroy(pacer);
	free(str_certificate);
	free(str_credetial);
	return 0;
}
